import pytest

from zoho_crm import ZohoAuthConfig, ZohoClient


@pytest.fixture
def zoho_client():
    return ZohoClient(
        zoho_auth_config=ZohoAuthConfig(
            client_id="clientid",
            client_secret="clientsecret",
            refresh_token="refreshtoken",
        )
    )



"Unoficial Zoho CRN API client"

__version__ = "0.0.4"

from .api_client import ZohoAPIConfig as ZohoAPIConfig
from .api_client import ZohoAuthConfig as ZohoAuthConfig
from .api_client import ZohoClient as ZohoClient
from .api_client import ZohoClientException as ZohoClientException
from .api_client import ZohoEndpoint as ZohoEndpoint



import logging
import os
from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, List, Tuple

import requests

logger = logging.getLogger(__name__)


class ZohoClientException(Exception):
    """handler errors from Zoho api"""


@dataclass
class ZohoAPIConfig:
    api_version: str = "v3"
    account_domain: str = "www.zohoapis.eu"
    api_results_per_page: int = 200
    api_discrete_pagination_limit: int = 2000

    @property
    def api_base_url(self) -> str:
        return f"https://{self.account_domain}/crm/{self.api_version}/"


@dataclass
class ZohoAuthConfig:
    client_id: str = os.environ.get("ZOHO_CLIENT_ID")
    client_secret: str = os.environ.get("ZOHO_CLIENT_SECRET")
    refresh_token: str = os.environ.get("ZOHO_REFRESH_TOKEN")
    auth_api_version: str = "v2"
    account_domain: str = "accounts.zoho.eu"

    @property
    def auth_base_url(self) -> str:
        return f"https://{self.account_domain}/oauth/{self.auth_api_version}/token"

    @property
    def refresh_access_token_url(self) -> str:
        return (
            f"{self.auth_base_url}?"
            f"refresh_token={self.refresh_token}&"
            f"client_id={self.client_id}&"
            f"client_secret={self.client_secret}&"
            f"grant_type=refresh_token"
        )


class FileTypes(Enum):
    PHOTO = "photo"
    ATTACHMENT = "Attachments"


@dataclass
class ZohoEndpoint:
    module_name: str
    method: str
    response_data_key: str = "data"
    file_type: FileTypes = None
    params: dict = field(default_factory=dict)
    data: dict = field(default_factory=dict)
    files: dict = field(default_factory=dict)

    @property
    def url(self):
        if "id" in self.params:
            record_url = f"{self.module_name}/{self.params['id']}"
            if self.file_type is not None:
                record_url = f"{record_url}/{self.file_type}"
            return record_url
        if "criteria" in self.params:
            return f"{self.module_name}/search"
        return self.module_name


class ZohoClient:
    def __init__(
        self,
        zoho_api_config: ZohoAPIConfig = None,
        zoho_auth_config: ZohoAuthConfig = None,
    ):
        self.zoho_config = zoho_api_config if zoho_api_config else ZohoAPIConfig()
        self.zoho_auth = zoho_auth_config if zoho_auth_config else ZohoAuthConfig()

    def _validate_response(self, response: Dict) -> Dict:
        if "error" in response or "error" == (
            response.get("status") or response.get("data", [{}])[0].get("status")
        ):
            raise ZohoClientException(f"The zoho server returned an error: {response}")
        return response

    def _get_new_access_token(self) -> str:
        """
        Renew the client access token and return the new one
        This action revoke the older ones
        """
        response_dict = requests.post(self.zoho_auth.refresh_access_token_url).json()
        response_dict = self._validate_response(response_dict)
        logger.debug("new access token got it: %s", response_dict)

        return response_dict["access_token"]

    @property
    def request_headers(self) -> str:
        try:
            return self._request_headers
        except AttributeError:
            token = self._get_new_access_token()
            self._request_headers = {"Authorization": f"Zoho-oauthtoken {token}"}
            return self._request_headers

    def _get_next_page_url(self, page_info: Dict, params: Dict) -> bool:
        if not page_info["more_records"]:
            return False

        amount = int(page_info["page"]) * page_info["per_page"]
        if amount >= self.zoho_config.api_discrete_pagination_limit:
            params["page_token"] = page_info["next_page_token"]
            params.pop("page", None)
            return True

        params["page"] = page_info["page"] + 1
        return True

    def _get_page(self, zoho_endpoint: ZohoEndpoint) -> Tuple[List[Dict], str | bool]:
        next_url = False
        try:
            response_dict = requests.get(
                self.zoho_config.api_base_url + zoho_endpoint.url,
                headers=self.request_headers,
                params=zoho_endpoint.params,
            ).json()
        except requests.exceptions.JSONDecodeError:
            return [], False

        valid_response_dict = self._validate_response(response_dict)
        if "info" in valid_response_dict:
            next_url = self._get_next_page_url(
                valid_response_dict["info"], zoho_endpoint.params
            )
        return valid_response_dict[zoho_endpoint.response_data_key], next_url

    def _call_read_endpoint(self, zoho_endpoint: ZohoEndpoint) -> List[Dict]:
        response_list = []

        next_page = True
        while next_page:
            result_list, next_page = self._get_page(zoho_endpoint)
            response_list.extend(result_list)

        return response_list

    def _call_write_endpoint(self, zoho_endpoint: ZohoEndpoint) -> Dict:
        try:
            response_dict = getattr(requests, zoho_endpoint.method)(
                self.zoho_config.api_base_url + zoho_endpoint.url,
                headers=self.request_headers,
                json=zoho_endpoint.data,
                files=zoho_endpoint.files,
            ).json()
        except requests.exceptions.JSONDecodeError:
            return {}
        valid_response_dict = self._validate_response(response_dict)
        return valid_response_dict

    def call(self, zoho_endpoint: ZohoEndpoint) -> Dict:
        if zoho_endpoint.method == "get":
            return self._call_read_endpoint(zoho_endpoint)
        return self._call_write_endpoint(zoho_endpoint)



import pytest
from pytest_mock import MockerFixture
from requests_mock import Mocker

from zoho_crm import (
    ZohoAPIConfig,
    ZohoAuthConfig,
    ZohoClient,
    ZohoClientException,
    ZohoEndpoint,
)


class TestZohoAPIConfig:
    def test_api_base_url(self):
        assert ZohoAPIConfig().api_base_url == "https://www.zohoapis.eu/crm/v3/"


class TestZohoAuthConfig:
    def test_auth_base_url(self):
        assert (
            ZohoAuthConfig().auth_base_url == "https://accounts.zoho.eu/oauth/v2/token"
        )

    def test_refresh_access_token(self):
        assert ZohoAuthConfig(
            client_id="12345", client_secret="6789", refresh_token="token1"
        ).refresh_access_token_url == (
            "https://accounts.zoho.eu/oauth/v2/"
            "token?refresh_token=token1&client_id=12345&client_secret=6789&grant_type=refresh_token"
        )


class TestZohoEndpoint:
    @pytest.mark.parametrize(
        "file_type,params,expected",
        [
            (None, {"1": 1}, "patata"),
            ("photo", {"id": 1}, "patata/1/photo"),
            ("Attachments", {"id": 1}, "patata/1/Attachments"),
            (None, {"criteria": "something"}, "patata/search"),
        ],
    )
    def test_url(self, file_type: str, params: dict, expected: str):
        assert (
            ZohoEndpoint(
                module_name="patata", method="get", params=params, file_type=file_type
            ).url
            == expected
        )


class TestZohoClient:
    def test__init__no_params(self):
        zoho_client = ZohoClient()

        assert isinstance(zoho_client.zoho_config, ZohoAPIConfig)
        assert isinstance(zoho_client.zoho_auth, ZohoAuthConfig)

    def test__init__(self):
        zoho_client = ZohoClient(
            zoho_api_config=ZohoAPIConfig(), zoho_auth_config=ZohoAuthConfig()
        )

        assert isinstance(zoho_client.zoho_config, ZohoAPIConfig)
        assert isinstance(zoho_client.zoho_auth, ZohoAuthConfig)

    def test_validate_response_ok(self, zoho_client: ZohoClient):
        assert zoho_client._validate_response({"patata": 1}) == {"patata": 1}

    @pytest.mark.parametrize(
        "response",
        [({"error": 1}), ({"status": "error"}), ({"data": [{"status": "error"}]})],
    )
    def test_validate_response_error(self, response: dict):
        with pytest.raises(ZohoClientException):
            ZohoClient()._validate_response(response)

    def test__get_new_access_token_KO(
        self, zoho_client: ZohoClient, requests_mock: Mocker
    ):
        requests_mock.post(
            zoho_client.zoho_auth.refresh_access_token_url, json={"error": "error 1"}
        )

        with pytest.raises(ZohoClientException):
            zoho_client._get_new_access_token()

    def test__get_new_access_token_OK(
        self, zoho_client: ZohoClient, requests_mock: Mocker
    ):
        requests_mock.post(
            zoho_client.zoho_auth.refresh_access_token_url,
            json={"access_token": "AA1234"},
        )

        assert zoho_client._get_new_access_token() == "AA1234"

    def test_request_headers_not_cached(
        self, zoho_client: ZohoClient, mocker: MockerFixture, requests_mock: Mocker
    ):
        requests_mock.post(
            zoho_client.zoho_auth.refresh_access_token_url,
            json={"access_token": "AA1234"},
        )
        spy__get_new_access_token = mocker.patch.object(
            zoho_client,
            "_get_new_access_token",
            wraps=zoho_client._get_new_access_token,
        )

        returned_headers = zoho_client.request_headers

        spy__get_new_access_token.assert_called_once()
        assert zoho_client._request_headers == {
            "Authorization": "Zoho-oauthtoken AA1234"
        }
        assert returned_headers == {"Authorization": "Zoho-oauthtoken AA1234"}

    def test_request_headers_already_cached(self, zoho_client: ZohoClient):
        zoho_client._request_headers = {"example": "example"}

        assert zoho_client.request_headers == {"example": "example"}

    def test__get_next_page_url_no_more(self, zoho_client: ZohoClient):
        page_info = {"more_records": False}
        next_page = zoho_client._get_next_page_url(page_info, {"page": 5})

        assert next_page is False

    def test__get_next_page_url_with_page_token(self, zoho_client: ZohoClient):
        params = {"page": 1}
        page_info = {
            "page": 3000,
            "more_records": True,
            "per_page": 2000,
            "next_page_token": "fakenextpagetoken1",
        }
        next_page = zoho_client._get_next_page_url(page_info, params)

        assert next_page is True
        assert "page" not in params
        assert params["page_token"] == "fakenextpagetoken1"

    def test__get_next_page_url_with_page_number(self, zoho_client: ZohoClient):
        params = {"page": 1}
        page_info = {"page": 1, "more_records": True, "per_page": 2}

        next_page = zoho_client._get_next_page_url(page_info, params)

        assert params["page"] == 2
        assert next_page

    def test__get_page_without_next_url(
        self, zoho_client: ZohoClient, requests_mock: Mocker
    ):
        zoho_endpoint = ZohoEndpoint(module_name="testmodule", method="get")
        zoho_client._request_headers = {"authorization": "fake"}
        fake_response = {
            "info": {"more_records": False},
            "data": [{"example": "example"}],
        }
        requests_mock.get(
            "https://www.zohoapis.eu/crm/v3/testmodule", json=fake_response
        )

        response_data, next_url = zoho_client._get_page(zoho_endpoint)

        assert response_data == [{"example": "example"}]
        assert not next_url

    def test__get_page_with_next_url(
        self, zoho_client: ZohoClient, requests_mock: Mocker
    ):
        zoho_endpoint = ZohoEndpoint(module_name="testmodule", method="get")
        zoho_client._request_headers = {"authorization": "fake"}
        requests_mock.get(
            "https://www.zohoapis.eu/crm/v3/testmodule",
            json={
                "info": {"more_records": True, "page": 1, "per_page": 2},
                "data": [{"example": "example"}],
            },
        )

        response_data, next_url = zoho_client._get_page(zoho_endpoint)

        assert response_data == [{"example": "example"}]
        assert next_url

    def test__get_page_json_decode_error(
        self, zoho_client: ZohoClient, requests_mock: MockerFixture
    ):
        zoho_endpoint = ZohoEndpoint(module_name="testmodule", method="get")
        zoho_client._request_headers = {"authorization": "fake"}
        requests_mock.get(
            "https://www.zohoapis.eu/crm/v3/testmodule",
            text="...",
        )

        response_data, next_url = zoho_client._get_page(zoho_endpoint)

        assert response_data == []
        assert not next_url

    def test__call_read_endpoint(self, zoho_client: ZohoClient, mocker: MockerFixture):
        mocker.patch(
            "zoho_crm.ZohoClient._get_page",
            side_effect=[([1, 2], True), ([3, 4], False)],
        )
        zoho_endpoint = ZohoEndpoint(module_name="testmodule", method="get")

        assert zoho_client._call_read_endpoint(zoho_endpoint) == [1, 2, 3, 4]

    def test__call_write_endpoint(self, zoho_client: ZohoClient, requests_mock: Mocker):
        requests_mock.post(
            "https://www.zohoapis.eu/crm/v3/testmodule",
            json={"response": "OK"},
        )
        zoho_client._request_headers = {"authorization": "fake"}
        zoho_endpoint = ZohoEndpoint(module_name="testmodule", method="post")

        assert zoho_client._call_write_endpoint(zoho_endpoint) == {"response": "OK"}

    def test__call_write_endpoint_json_decode_error(
        self, zoho_client: ZohoClient, requests_mock: Mocker
    ):
        requests_mock.post(
            "https://www.zohoapis.eu/crm/v3/testmodule",
            text="...",
        )
        zoho_client._request_headers = {"authorization": "fake"}
        zoho_endpoint = ZohoEndpoint(module_name="testmodule", method="post")

        assert zoho_client._call_write_endpoint(zoho_endpoint) == {}

    @pytest.mark.parametrize(
        "mock_address,method",
        [
            ("zoho_crm.ZohoClient._call_read_endpoint", "get"),
            ("zoho_crm.ZohoClient._call_write_endpoint", "post"),
        ],
    )
    def test_call(
        self,
        zoho_client: ZohoClient,
        mocker: MockerFixture,
        mock_address: str,
        method: str,
    ):
        mock_endpoint = mocker.patch(mock_address)
        zoho_endpoint = ZohoEndpoint(module_name="testmodule", method=method)

        zoho_client.call(zoho_endpoint)

        mock_endpoint.assert_called_with(zoho_endpoint)
